"""Constants that define defaults."""
import re

EXCLUDE = (
    '.svn',
    'CVS',
    '.bzr',
    '.hg',
    '.git',
    '__pycache__',
    '.tox',
    '.eggs',
    '*.egg',
)
IGNORE = (
    'E121',
    'E123',
    'E126',
    'E226',
    'E24',
    'E704',
    'W503',
    'W504',
)
SELECT = ('E', 'F', 'W', 'C90')
MAX_LINE_LENGTH = 79

TRUTHY_VALUES = {'true', '1', 't'}

# Other constants
WHITESPACE = frozenset(' \t')

STATISTIC_NAMES = (
    'logical lines',
    'physical lines',
    'tokens',
)

NOQA_INLINE_REGEXP = re.compile(
    # We're looking for items that look like this:
    # ``# noqa``
    # ``# noqa: E123``
    # ``# noqa: E123,W451,F921``
    # ``# NoQA: E123,W451,F921``
    # ``# NOQA: E123,W451,F921``
    # We do not care about the ``: `` that follows ``noqa``
    # We do not care about the casing of ``noqa``
    # We want a comma-separated list of errors
    '# noqa(?:: (?P<codes>([A-Z][0-9]+(?:[,\s]+)?)+))?',
    re.IGNORECASE
)

NOQA_FILE = re.compile(r'\s*# flake8[:=]\s*noqa', re.I)



"""Config handling logic for Flake8."""
import collections
import configparser
import logging
import os.path
import sys

from flake8 import utils

LOG = logging.getLogger(__name__)

__all__ = ('ConfigFileFinder', 'MergedConfigParser')


class ConfigFileFinder(object):
    """Encapsulate the logic for finding and reading config files."""

    PROJECT_FILENAMES = ('setup.cfg', 'tox.ini')

    def __init__(self, program_name, args, extra_config_files):
        """Initialize object to find config files.

        :param str program_name:
            Name of the current program (e.g., flake8).
        :param list args:
            The extra arguments passed on the command-line.
        :param list extra_config_files:
            Extra configuration files specified by the user to read.
        """
        # The values of --append-config from the CLI
        extra_config_files = extra_config_files or []
        self.extra_config_files = [
            # Ensure the paths are absolute paths for local_config_files
            os.path.abspath(f) for f in extra_config_files
        ]

        # Platform specific settings
        self.is_windows = sys.platform == 'win32'
        self.xdg_home = os.environ.get('XDG_CONFIG_HOME',
                                       os.path.expanduser('~/.config'))

        # Look for '.<program_name>' files
        self.program_config = '.' + program_name
        self.program_name = program_name

        # List of filenames to find in the local/project directory
        self.project_filenames = ('setup.cfg', 'tox.ini', self.program_config)

        self.local_directory = os.path.abspath(os.curdir)

        if not args:
            args = ['.']
        self.parent = self.tail = os.path.abspath(os.path.commonprefix(args))

        # caches to avoid double-reading config files
        self._local_configs = None
        self._user_config = None
        self._cli_configs = {}

    @staticmethod
    def _read_config(files):
        config = configparser.RawConfigParser()
        if isinstance(files, (str, type(u''))):
            files = [files]

        found_files = []
        for filename in files:
            try:
                found_files.extend(config.read(filename))
            except UnicodeDecodeError:
                LOG.exception("There was an error decoding a config file."
                              "The file with a problem was %s.",
                              filename)
            except configparser.ParsingError:
                LOG.exception("There was an error trying to parse a config "
                              "file. The file with a problem was %s.",
                              filename)
        return (config, found_files)

    def cli_config(self, files):
        """Read and parse the config file specified on the command-line."""
        if files not in self._cli_configs:
            config, found_files = self._read_config(files)
            if found_files:
                LOG.debug('Found cli configuration files: %s', found_files)
            self._cli_configs[files] = config
        return self._cli_configs[files]

    def generate_possible_local_files(self):
        """Find and generate all local config files."""
        tail = self.tail
        parent = self.parent
        found_config_files = False
        while tail and not found_config_files:
            for project_filename in self.project_filenames:
                filename = os.path.abspath(os.path.join(parent,
                                                        project_filename))
                if os.path.exists(filename):
                    yield filename
                    found_config_files = True
                    self.local_directory = parent
            (parent, tail) = os.path.split(parent)

    def local_config_files(self):
        """Find all local config files which actually exist.

        Filter results from
        :meth:`~ConfigFileFinder.generate_possible_local_files` based
        on whether the filename exists or not.

        :returns:
            List of files that exist that are local project config files with
            extra config files appended to that list (which also exist).
        :rtype:
            [str]
        """
        exists = os.path.exists
        return [
            filename
            for filename in self.generate_possible_local_files()
        ] + [f for f in self.extra_config_files if exists(f)]

    def local_configs(self):
        """Parse all local config files into one config object."""
        if self._local_configs is None:
            config, found_files = self._read_config(self.local_config_files())
            if found_files:
                LOG.debug('Found local configuration files: %s', found_files)
            self._local_configs = config
        return self._local_configs

    def user_config_file(self):
        """Find the user-level config file."""
        if self.is_windows:
            return os.path.expanduser('~\\' + self.program_config)
        return os.path.join(self.xdg_home, self.program_name)

    def user_config(self):
        """Parse the user config file into a config object."""
        if self._user_config is None:
            config, found_files = self._read_config(self.user_config_file())
            if found_files:
                LOG.debug('Found user configuration files: %s', found_files)
            self._user_config = config
        return self._user_config


class MergedConfigParser(object):
    """Encapsulate merging different types of configuration files.

    This parses out the options registered that were specified in the
    configuration files, handles extra configuration files, and returns
    dictionaries with the parsed values.
    """

    #: Set of types that should use the
    #: :meth:`~configparser.RawConfigParser.getint` method.
    GETINT_TYPES = {'int', 'count'}
    #: Set of actions that should use the
    #: :meth:`~configparser.RawConfigParser.getbool` method.
    GETBOOL_ACTIONS = {'store_true', 'store_false'}

    def __init__(self, option_manager, config_finder):
        """Initialize the MergedConfigParser instance.

        :param flake8.options.manager.OptionManager option_manager:
            Initialized OptionManager.
        :param flake8.options.config.ConfigFileFinder config_finder:
            Initialized ConfigFileFinder.
        """
        #: Our instance of flake8.options.manager.OptionManager
        self.option_manager = option_manager
        #: The prog value for the cli parser
        self.program_name = option_manager.program_name
        #: Mapping of configuration option names to
        #: :class:`~flake8.options.manager.Option` instances
        self.config_options = option_manager.config_options_dict
        #: Our instance of our :class:`~ConfigFileFinder`
        self.config_finder = config_finder

    def _normalize_value(self, option, value):
        final_value = option.normalize(
            value,
            self.config_finder.local_directory,
        )
        LOG.debug('%r has been normalized to %r for option "%s"',
                  value, final_value, option.config_name)
        return final_value

    def _parse_config(self, config_parser):
        config_dict = {}
        for option_name in config_parser.options(self.program_name):
            if option_name not in self.config_options:
                LOG.debug('Option "%s" is not registered. Ignoring.',
                          option_name)
                continue
            option = self.config_options[option_name]

            # Use the appropriate method to parse the config value
            method = config_parser.get
            if (option.type in self.GETINT_TYPES or
                    option.action in self.GETINT_TYPES):
                method = config_parser.getint
            elif option.action in self.GETBOOL_ACTIONS:
                method = config_parser.getboolean

            value = method(self.program_name, option_name)
            LOG.debug('Option "%s" returned value: %r', option_name, value)

            final_value = self._normalize_value(option, value)
            config_dict[option.config_name] = final_value

        return config_dict

    def is_configured_by(self, config):
        """Check if the specified config parser has an appropriate section."""
        return config.has_section(self.program_name)

    def parse_local_config(self):
        """Parse and return the local configuration files."""
        config = self.config_finder.local_configs()
        if not self.is_configured_by(config):
            LOG.debug('Local configuration files have no %s section',
                      self.program_name)
            return {}

        LOG.debug('Parsing local configuration files.')
        return self._parse_config(config)

    def parse_user_config(self):
        """Parse and return the user configuration files."""
        config = self.config_finder.user_config()
        if not self.is_configured_by(config):
            LOG.debug('User configuration files have no %s section',
                      self.program_name)
            return {}

        LOG.debug('Parsing user configuration files.')
        return self._parse_config(config)

    def parse_cli_config(self, config_path):
        """Parse and return the file specified by --config."""
        config = self.config_finder.cli_config(config_path)
        if not self.is_configured_by(config):
            LOG.debug('CLI configuration files have no %s section',
                      self.program_name)
            return {}

        LOG.debug('Parsing CLI configuration files.')
        return self._parse_config(config)

    def merge_user_and_local_config(self):
        """Merge the parsed user and local configuration files.

        :returns:
            Dictionary of the parsed and merged configuration options.
        :rtype:
            dict
        """
        user_config = self.parse_user_config()
        config = self.parse_local_config()

        for option, value in user_config.items():
            config.setdefault(option, value)

        return config

    def parse(self, cli_config=None, isolated=False):
        """Parse and return the local and user config files.

        First this copies over the parsed local configuration and then
        iterates over the options in the user configuration and sets them if
        they were not set by the local configuration file.

        :param str cli_config:
            Value of --config when specified at the command-line. Overrides
            all other config files.
        :param bool isolated:
            Determines if we should parse configuration files at all or not.
            If running in isolated mode, we ignore all configuration files
        :returns:
            Dictionary of parsed configuration options
        :rtype:
            dict
        """
        if isolated:
            LOG.debug('Refusing to parse configuration files due to user-'
                      'requested isolation')
            return {}

        if cli_config:
            LOG.debug('Ignoring user and locally found configuration files. '
                      'Reading only configuration from "%s" specified via '
                      '--config by the user', cli_config)
            return self.parse_cli_config(cli_config)

        return self.merge_user_and_local_config()


def get_local_plugins(config_finder, cli_config=None, isolated=False):
    """Get local plugins lists from config files.

    :param flake8.options.config.ConfigFileFinder config_finder:
        The config file finder to use.
    :param str cli_config:
        Value of --config when specified at the command-line. Overrides
        all other config files.
    :param bool isolated:
        Determines if we should parse configuration files at all or not.
        If running in isolated mode, we ignore all configuration files
    :returns:
        LocalPlugins namedtuple containing two lists of plugin strings,
        one for extension (checker) plugins and one for report plugins.
    :rtype:
        flake8.options.config.LocalPlugins
    """
    local_plugins = LocalPlugins(extension=[], report=[])
    if isolated:
        LOG.debug('Refusing to look for local plugins in configuration'
                  'files due to user-requested isolation')
        return local_plugins

    if cli_config:
        LOG.debug('Reading local plugins only from "%s" specified via '
                  '--config by the user', cli_config)
        config = config_finder.cli_config(cli_config)
    else:
        config = config_finder.local_configs()

    section = '%s:local-plugins' % config_finder.program_name
    for plugin_type in ['extension', 'report']:
        if config.has_option(section, plugin_type):
            local_plugins_string = config.get(section, plugin_type).strip()
            plugin_type_list = getattr(local_plugins, plugin_type)
            plugin_type_list.extend(utils.parse_comma_separated_list(
                local_plugins_string,
                regexp=utils.LOCAL_PLUGIN_LIST_RE,
            ))
    return local_plugins


LocalPlugins = collections.namedtuple('LocalPlugins', 'extension report')



"""Unit test for the _trie module."""
from flake8.plugins import _trie as trie


class TestTrie(object):
    """Collection of tests for the Trie class."""

    def test_traverse_without_data(self):
        """Verify the behaviour when traversing an empty Trie."""
        tree = trie.Trie()
        assert list(tree.traverse()) == []

    def test_traverse_with_data(self):
        """Verify that traversal of our Trie is depth-first and pre-order."""
        tree = trie.Trie()
        tree.add('A', 'A')
        tree.add('a', 'a')
        tree.add('AB', 'B')
        tree.add('Ab', 'b')
        tree.add('AbC', 'C')
        tree.add('Abc', 'c')
        # The trie tree here should look something like
        #
        #      <root>
        #     /      \
        #    A        a
        #  /  |
        # B   b
        #    / \
        #   C   c
        #
        # And the traversal should look like:
        #
        # A B b C c a
        expected_order = ['A', 'B', 'b', 'C', 'c', 'a']
        for expected, actual_node in zip(expected_order, tree.traverse()):
            assert actual_node.prefix == expected

    def test_find(self):
        """Exercise the Trie.find method."""
        tree = trie.Trie()
        tree.add('A', 'A')
        tree.add('a', 'a')
        tree.add('AB', 'AB')
        tree.add('Ab', 'Ab')
        tree.add('AbC', 'AbC')
        tree.add('Abc', 'Abc')

        assert tree.find('AB').data == ['AB']
        assert tree.find('AbC').data == ['AbC']
        assert tree.find('A').data == ['A']
        assert tree.find('X') is None


class TestTrieNode(object):
    """Collection of tests for the TrieNode class."""

    def test_add_child(self):
        """Verify we add children appropriately."""
        node = trie.TrieNode('E', 'E is for Eat')
        assert node.find_prefix('a') is None
        added = node.add_child('a', 'a is for Apple')
        assert node.find_prefix('a') is added

    def test_add_child_overrides_previous_child(self):
        """Verify adding a child will replace the previous child."""
        node = trie.TrieNode('E', 'E is for Eat', children={
            'a': trie.TrieNode('a', 'a is for Apple')
        })
        previous = node.find_prefix('a')
        assert previous is not None

        added = node.add_child('a', 'a is for Ascertain')
        assert node.find_prefix('a') is added

    def test_find_prefix(self):
        """Verify we can find a child with the specified prefix."""
        node = trie.TrieNode('E', 'E is for Eat', children={
            'a': trie.TrieNode('a', 'a is for Apple')
        })
        child = node.find_prefix('a')
        assert child is not None
        assert child.prefix == 'a'
        assert child.data == 'a is for Apple'

    def test_find_prefix_returns_None_when_no_children_have_the_prefix(self):
        """Verify we receive None from find_prefix for missing children."""
        node = trie.TrieNode('E', 'E is for Eat', children={
            'a': trie.TrieNode('a', 'a is for Apple')
        })
        assert node.find_prefix('b') is None

    def test_traverse_does_nothing_when_a_node_has_no_children(self):
        """Verify traversing a node with no children does nothing."""
        node = trie.TrieNode('E', 'E is for Eat')
        assert list(node.traverse()) == []

    def test_traverse(self):
        """Verify traversal is depth-first and pre-order."""
        root = trie.TrieNode(None, None)
        node = root.add_child('A', 'A')
        root.add_child('a', 'a')
        node.add_child('B', 'B')
        node = node.add_child('b', 'b')
        node.add_child('C', 'C')
        node.add_child('c', 'c')
        # The sub-tree here should look something like
        #
        #      <root>
        #     /      \
        #    A        a
        #  /  |
        # B   b
        #    / \
        #   C   c
        #
        # And the traversal should look like:
        #
        # A B b C c a
        expected_order = ['A', 'B', 'b', 'C', 'c', 'a']
        for expected, actual_node in zip(expected_order, root.traverse()):
            assert actual_node.prefix == expected



"""Exception classes for all of Flake8."""


class Flake8Exception(Exception):
    """Plain Flake8 exception."""

    pass


class EarlyQuit(Flake8Exception):
    """Except raised when encountering a KeyboardInterrupt."""

    pass


class ExecutionError(Flake8Exception):
    """Exception raised during execution of Flake8."""


class FailedToLoadPlugin(Flake8Exception):
    """Exception raised when a plugin fails to load."""

    FORMAT = 'Flake8 failed to load plugin "%(name)s" due to %(exc)s.'

    def __init__(self, *args, **kwargs):
        """Initialize our FailedToLoadPlugin exception."""
        self.plugin = kwargs.pop('plugin')
        self.ep_name = self.plugin.name
        self.original_exception = kwargs.pop('exception')
        super(FailedToLoadPlugin, self).__init__(*args, **kwargs)

    def __str__(self):
        """Return a nice string for our exception."""
        return self.FORMAT % {'name': self.ep_name,
                              'exc': self.original_exception}


class InvalidSyntax(Flake8Exception):
    """Exception raised when tokenizing a file fails."""

    def __init__(self, *args, **kwargs):
        """Initialize our InvalidSyntax exception."""
        exception = kwargs.pop('exception', None)
        self.original_exception = exception
        self.error_message = '{0}: {1}'.format(
            exception.__class__.__name__,
            exception.args[0],
        )
        self.error_code = 'E902'
        self.line_number = 1
        self.column_number = 0
        super(InvalidSyntax, self).__init__(
            self.error_message,
            *args,
            **kwargs
        )


class PluginRequestedUnknownParameters(Flake8Exception):
    """The plugin requested unknown parameters."""

    FORMAT = '"%(name)s" requested unknown parameters causing %(exc)s'

    def __init__(self, *args, **kwargs):
        """Pop certain keyword arguments for initialization."""
        self.original_exception = kwargs.pop('exception')
        self.plugin = kwargs.pop('plugin')
        super(PluginRequestedUnknownParameters, self).__init__(
            *args,
            **kwargs
        )

    def __str__(self):
        """Format our exception message."""
        return self.FORMAT % {'name': self.plugin['plugin_name'],
                              'exc': self.original_exception}


class HookInstallationError(Flake8Exception):
    """Parent exception for all hooks errors."""

    pass


class GitHookAlreadyExists(HookInstallationError):
    """Exception raised when the git pre-commit hook file already exists."""

    def __init__(self, *args, **kwargs):
        """Initialize the path attribute."""
        self.path = kwargs.pop('path')
        super(GitHookAlreadyExists, self).__init__(*args, **kwargs)

    def __str__(self):
        """Provide a nice message regarding the exception."""
        msg = ('The Git pre-commit hook ({0}) already exists. To convince '
               'Flake8 to install the hook, please remove the existing '
               'hook.')
        return msg.format(self.path)


class MercurialHookAlreadyExists(HookInstallationError):
    """Exception raised when a mercurial hook is already configured."""

    hook_name = None

    def __init__(self, *args, **kwargs):
        """Initialize the relevant attributes."""
        self.path = kwargs.pop('path')
        self.value = kwargs.pop('value')
        super(MercurialHookAlreadyExists, self).__init__(*args, **kwargs)

    def __str__(self):
        """Return a nicely formatted string for these errors."""
        msg = ('The Mercurial {0} hook already exists with "{1}" in {2}. '
               'To convince Flake8 to install the hook, please remove the '
               '{0} configuration from the [hooks] section of your hgrc.')
        return msg.format(self.hook_name, self.value, self.path)


class MercurialCommitHookAlreadyExists(MercurialHookAlreadyExists):
    """Exception raised when the hg commit hook is already configured."""

    hook_name = 'commit'


class MercurialQRefreshHookAlreadyExists(MercurialHookAlreadyExists):
    """Exception raised when the hg commit hook is already configured."""

    hook_name = 'qrefresh'



"""Tests for the flake8.style_guide.Violation class."""
import mock
import pytest

from flake8 import style_guide


@pytest.mark.parametrize('error_code,physical_line,expected_result', [
    ('E111', 'a = 1', False),
    ('E121', 'a = 1  # noqa: E111', False),
    ('E121', 'a = 1  # noqa: E111,W123,F821', False),
    ('E111', 'a = 1  # noqa: E111,W123,F821', True),
    ('W123', 'a = 1  # noqa: E111,W123,F821', True),
    ('W123', 'a = 1  # noqa: E111, W123,F821', True),
    ('E111', 'a = 1  # noqa: E11,W123,F821', True),
    ('E111', 'a = 1  # noqa, analysis:ignore', True),
    ('E111', 'a = 1  # noqa analysis:ignore', True),
    ('E111', 'a = 1  # noqa - We do not care', True),
    ('E111', 'a = 1  # noqa: We do not care', True),
])
def test_is_inline_ignored(error_code, physical_line, expected_result):
    """Verify that we detect inline usage of ``# noqa``."""
    error = style_guide.Violation(
        error_code, 'filename.py', 1, 1, 'error text', None)
    # We want `None` to be passed as the physical line so we actually use our
    # monkey-patched linecache.getline value.

    with mock.patch('linecache.getline', return_value=physical_line):
        assert error.is_inline_ignored(False) is expected_result


def test_disable_is_inline_ignored():
    """Verify that is_inline_ignored exits immediately if disabling NoQA."""
    error = style_guide.Violation(
        'E121', 'filename.py', 1, 1, 'error text', 'line')

    with mock.patch('linecache.getline') as getline:
        assert error.is_inline_ignored(True) is False

    assert getline.called is False


@pytest.mark.parametrize('violation_file,violation_line,diff,expected', [
    ('file.py', 10, {}, True),
    ('file.py', 1, {'file.py': range(1, 2)}, True),
    ('file.py', 10, {'file.py': range(1, 2)}, False),
    ('file.py', 1, {'other.py': range(1, 2)}, False),
    ('file.py', 10, {'other.py': range(1, 2)}, False),
])
def test_violation_is_in_diff(violation_file, violation_line, diff, expected):
    """Verify that we find violations within a diff."""
    violation = style_guide.Violation(
        'E001', violation_file, violation_line, 1, 'warning', 'line',
    )

    assert violation.is_in(diff) is expected
